from typing import List
import pandas as pd
from pandas.core.frame import DataFrame
from sklearn.preprocessing import MinMaxScaler


def get_low_pearson_r(
    df_well: DataFrame, r_window_size: int, logname: str, col: str, thresh: float
) -> DataFrame:
    """
    Flags samples with low correlation compared to window around it

    Args:
        df_well (DataFrame): data of one well
        r_window_size (int): window size
        logname (str): log being analyzed
        col (str): column being compared to
        thresh (float): threshold for being a bad correlation

    Returns:
        DataFrame: flagged dataframe
    """
    df_interpolated = df_well[[logname, col]].interpolate()
    rolling_r = df_interpolated[logname].rolling(
        window=r_window_size,
        center=True
    ).corr(df_interpolated[col])
    df_well[f"r_{col}"] = rolling_r
    df_well[f"bad_wr_{col}"] = df_well[f"r_{col}"] < thresh
    # Fill the gap between bad samples if two consecutive bad samples are less
    # than "r_window_size" apart
    mask = df_well[f"bad_wr_{col}"].groupby(
        (
            df_well[f"bad_wr_{col}"] != df_well[f"bad_wr_{col}"].shift()
        ).cumsum()
    ).transform("count").lt(r_window_size)
    mask &= df_well[f"bad_wr_{col}"].eq(False)
    df_well[f"bad_wr_{col}"].update(
        df_well[f"bad_wr_{col}"].loc[mask].replace(False, True)
    )
    return df_well


def flag_well(
    df_well: DataFrame, logname: str, cols: List, r_window_size: int, thresh: float
) -> DataFrame:
    """
    Flags samples with anomalous correlations

    Args:
        df_well (DataFrame): data of one well
        logname (str): log being analyzed
        cols (list): list of columns to compare logname to
        r_window_size (int): window size
        thresh (float): window value threshold to which consider sample an anomaly

    Returns:
        DataFrame: flagged dataframe
    """
    df_well_flagged = pd.DataFrame()
    df_well[[logname]] = MinMaxScaler().fit_transform(df_well[[logname]].values)
    df_well["bad_r"] = False
    for col in cols:
        df_well[[col]] = MinMaxScaler().fit_transform(df_well[[logname]].values)
        df_well = get_low_pearson_r(df_well, r_window_size, logname, col, thresh)
        df_well["bad_r"] = df_well[f"bad_wr_{col}"] | df_well["bad_r"]
    df_well_flagged = df_well_flagged.append(df_well)
    return df_well_flagged


def get_corr_anomalies(
    df_well: DataFrame, logname: str, cols: List, r_window_size: int, thresh: float
) -> List:
    """
    Flags correlations anomalies

    Args:
        df_well (DataFrame): data of one well
        logname (str): log being analyzed
        cols (list): list of columns to compare logname to
        r_window_size (int): window size
        thresh ([type]): [description]

    Returns:
        list: list of anomalous indices
    """
    df_well.fillna(method="ffill", inplace=True)
    df_well.fillna(method="bfill", inplace=True)
    df_flagged = flag_well(
        df_well, logname,
        [c for c in cols if c != logname],
        r_window_size,
        thresh
    )
    df_flagged.fillna(False, inplace=True)
    return df_flagged[df_flagged.bad_r == True].index


def flag_correlate(
    df_well: DataFrame, y_pred: DataFrame = None, **kwargs
) -> DataFrame:
    """
    Returns anomalous correlation indices for AC, ACS and DEN

    Args:
        df_well (DataFrame): df with one wells data
        y_pred (DataFrame, optional): df with results of predictions. Defaults to None.
    Returns:
        DataFrame: df with flag columns
    """
    print("Method: correlation...")
    if y_pred is None:
        y_pred = df_well.copy()
    expected_curves = kwargs.get(
        "expected_curves",
        set([
            "DEPTH", "DEN", "DENC", "AC", "ACS", "BS", "CALI",
            "GR", "NEU", "RDEP", "RMED", "RMIC", "GROUP"
        ])
    )
    y_pred.loc[:, [
        "flag_correlation_gen",
        "flag_correlation_ac",
        "flag_correlation_acs",
        "flag_correlation_den"
    ]] = 0, 0, 0, 0
    r_window_size = kwargs.get("r_window_size", 5)
    thresh = kwargs.get("thresh", 1)
    cols = kwargs.get("cols", list(expected_curves))
    ac_corr_anomalies = get_corr_anomalies(
        df_well, "AC", cols, r_window_size, thresh
    )
    acs_corr_anomalies = get_corr_anomalies(
        df_well, "ACS", cols, r_window_size, thresh
    )
    den_corr_anomalies = get_corr_anomalies(
        df_well, "DEN", cols, r_window_size, thresh
    )
    y_pred.loc[
        ac_corr_anomalies, ["flag_correlation_gen", "flag_correlation_ac"]
    ] = 1
    y_pred.loc[
        acs_corr_anomalies, ["flag_correlation_gen", "flag_correlation_acs"]
    ] = 1
    y_pred.loc[
        den_corr_anomalies, ["flag_correlation_gen", "flag_correlation_den"]
    ] = 1
    return y_pred



from typing import List
import numpy as np
from pandas.core.frame import DataFrame
from akerbp.models.rule_based_models import BS_helpers


def flag_bad_den_shallow(df_well: DataFrame) -> List[int]:
    """
    Looks for parial DEN measurement for the same BS
    in the shallow end of the well and mark those DEN measurements as bad

    Args:
        df_well (DataFrame): well data

    Returns:
        list: indices with bad DEN
    """
    df_well = BS_helpers.find_BS_jumps(df_well)
    dencorr_casing = []
    # get the index of the first DEN value
    first_DEN_idx = df_well[df_well["DEN"].notna()].index.min()
    if not np.isnan(first_DEN_idx):
        first_BS_region = df_well.loc[first_DEN_idx, "BS_region"]
        first_region_idx = df_well[df_well["BS_region"] == first_BS_region].index.min()
        bad_region_length = len(
            df_well[df_well["BS_region"] == first_BS_region]["DEN"].dropna()
        )
        if (first_region_idx < first_DEN_idx) and (bad_region_length < 1000):
            dencorr_casing = df_well[df_well["BS_region"] == first_BS_region].index
    return dencorr_casing


def flag_bad_den_at_jump(df_well: DataFrame) -> List[int]:
    """
    Looks around the regions where BS value changes, if there are missing DEN
    measurements, marks X measurements before and after that as potentioally bad

    Args:
        df_well (DataFrame): well data

    Returns:
        list: indices with bad DEN
    """
    df_well = BS_helpers.find_BS_jumps(df_well)
    bad_den_idx = []
    for region in df_well["BS_region"].unique():
        min_region_idx = df_well[df_well["BS_region"] == region].index.min()
        min_den_idx = df_well[
            df_well["BS_region"] == region
        ]["DEN"].dropna().index.min()
        if min_den_idx > min_region_idx:
            bad_den_idx.extend(df_well[
                df_well["BS_region"] == region]["DEN"].dropna().head(20).index.tolist()
            )
        max_region_idx = df_well[df_well["BS_region"] == region].index.max()
        max_den_idx = df_well[
            df_well["BS_region"] == region
        ]["DEN"].dropna().index.max()
        if max_den_idx < max_region_idx:
            bad_den_idx.extend(df_well[
                df_well["BS_region"] == region]["DEN"].dropna().tail(20).index.tolist()
            )
    return bad_den_idx


def flag_casing(df_well: DataFrame, y_pred: DataFrame = None, **kwargs) -> DataFrame:
    """
    Returns anomalous DEN due to invalid measurements inside the casing
    combining the findings of flag_bad_den_shallow() and flag_bad_den_at_jump()
    function

    Args:
        df_well (DataFrame): well data
        y_pred (DataFrame): dataframe to append the flag columns
        ("flag_casing_gen" and "flag_casing_den") to.
        If set to None, the columns are added to a copy of df_well - default: None

    Returns:
        DataFrame: y_pred
    """
    print("Method: casing...")
    bad_den_casing = []
    bad_den_casing.extend(flag_bad_den_shallow(df_well))
    bad_den_casing.extend(flag_bad_den_at_jump(df_well))
    if y_pred is None:
        y_pred = df_well.copy()
    y_pred.loc[:, ["flag_casing_gen", "flag_casing_den"]] = 0, 0
    y_pred.loc[bad_den_casing, ["flag_casing_gen", "flag_casing_den"]] = 1
    return y_pred



import numpy as np
from pandas.core.frame import DataFrame


def _number_BS_regions(df_: DataFrame) -> DataFrame:
    """
    Gives each BS region a number

    Args:
        df (DataFrame): should include "BS_jump" column (others will be ignored)

    Returns:
        df (DataFrame): same as the input dataframe with the added "BS_region"
        column with dtype=int
    """
    df = df_.copy()
    start_idx = df.index.min()
    df["BS_region"] = np.nan
    for rid, idx in enumerate(df[df["BS_jump"] == 1].index):
        end_idx = idx
        df.loc[range(start_idx, end_idx), "BS_region"] = rid
        start_idx = end_idx
    df["BS_region"].fillna(0, inplace=True)
    return df


def find_BS_jumps(df_: DataFrame) -> DataFrame:
    """
    Finds points where BS changes

    Args:
        df (DataFrame): should include "BS" column (others will be ignored)

    Returns:
        df (DataFrame): same as the input dataframe, with the added "BS_jump"
        with dtype="bool"
    """
    df = df_.copy()
    # First find points where BS changes
    df["BS_jump"] = 0
    df.loc[df["BS"].diff(1).ne(0), "BS_jump"] = 1
    # mark each BS region with an ID
    df = _number_BS_regions(df)
    return df


def flag_BS(
    df_well: DataFrame, y_pred: DataFrame = None, **kwargs
) -> DataFrame:
    """
    Returns anomalous BS values

    Args:
        df_well (DataFrame): dataframe to flag BS

    Returns:
        df (DataFrame): dataframe with results of BS flagging
    """
    print("Method: bitsize...")
    bs_step_size = kwargs.get("bs_step_size", 10)
    df_well = find_BS_jumps(df_well)
    bs_anomalies = []
    for _, v in df_well[
        df_well["BS_jump"] == 0
    ].groupby((df_well["BS_jump"] != 0).cumsum()):
        if v.shape[0] < bs_step_size:
            bs_anomalies.extend(v.index.tolist())
    if y_pred is None:
        y_pred = df_well.copy()
    y_pred.loc[:, ["flag_bitsize_gen"]] = 0
    y_pred.loc[bs_anomalies, ["flag_bitsize_gen"]] = 1
    return y_pred



import numpy as np
from pandas.core.frame import DataFrame
from akerbp.models.rule_based_models import helpers
from akerbp.models.rule_based_models import BS_helpers
from pyod.utils.utility import standardizer


def get_outliers(
    df_well: DataFrame, y_pred: DataFrame = None
) -> DataFrame:
    """
    Returns CALI-BS outliers

    Args:
        df_well (DataFrame): well data
        y_pred (DataFrame): dataframe with flags

    Returns:
        DataFrame: y_pred with added flag columns
    """
    df_well.loc[:, ["flag_washout_gen", "flag_washout_den"]] = 0, 0
    if y_pred is None:
        y_pred = df_well.copy()

    if "CALI" in df_well.columns and "BS" in df_well.columns:
        df_well.loc[:, "CALI_BS"] = np.abs(df_well["CALI"] - df_well["BS"])
    else:
        return y_pred

    x = "CALI_BS"
    y = "CALI_BS"
    algo_params = {
        "DBSCAN_eps": 0.02,
        "EliEnv_contamination": 0.01,
        "EliEnv_random_state": 0,
        "SVM_gamma": "scale",
        "SVM_nu": 0.01,
        "IsoFor_n_estimators": 20,
        "IsoFor_contamination": 0.05,
        "IsoFor_random_state": 0
    }

    df_well = BS_helpers.find_BS_jumps(df_well)

    # return if too many BS regions (a couple of wells with non-standard values)
    if df_well["BS_region"].nunique() >= 10:
        return df_well

    den_washout_anomalies = []

    for bsr in df_well["BS_region"].unique():
        df_gr = df_well[df_well["BS_region"] == bsr].copy()
        if (df_gr[x].dropna().shape[0] != 0) and (df_gr[y].dropna().shape[0] != 0):
            y_scores = helpers.get_crossplot_scores(
                df_gr, "CALI_BS", ["CALI_BS"], y_pred, "washout", **algo_params
            )
            y_pred[y_scores.columns] = y_scores

            den_washout_anomalies.extend(
                helpers.get_crossplot_anomalies(
                    df_gr, "CALI_BS", ["CALI_BS"], **algo_params
                )
            )
    y_pred.loc[den_washout_anomalies, ["flag_washout_gen", "flag_washout_den"]] = 1

    return y_pred


def flag_washout(df_well: DataFrame, y_pred: DataFrame = None, **kwargs) -> DataFrame:
    """
    Returns anomalous CALI-BS

    Args:
        df_well (DataFrame): well data
        y_pred (DataFrame): dataframe to add detection columns to

    Returns:
        DataFrame: y_pred with the added columns
    """
    print("Method: washout...")
    y_pred = get_outliers(df_well, y_pred)
    if "score_washout_gen" not in y_pred.columns:
        y_pred.loc[:, "score_washout_gen"] = 0
    rolling_window = kwargs.get("washout_rolling_window", 101)
    if not "BS_region" in y_pred.columns:
        return y_pred
    for bsr in y_pred["BS_region"].unique():
        df_gr = y_pred[y_pred["BS_region"] == bsr].copy()
        rows_with_values = df_gr[df_gr["CALI"].isna() == False].index
        if len(rows_with_values):
            df_gr.loc[rows_with_values, "CALI_std"] =\
                df_gr.loc[
                    rows_with_values, "CALI"].rolling(rolling_window).std().fillna(0)
            y_pred.loc[rows_with_values, "score_washout_gen"] =\
                np.abs(standardizer(
                    df_gr.loc[rows_with_values, "CALI_std"].values.reshape(-1, 1)
                ))
    y_pred["score_washout_gen"].fillna(0, inplace=True)
    return y_pred



from typing import List, Tuple
from pandas.core.frame import DataFrame, Series
import numpy as np
import sklearn
import matplotlib.pyplot as plt
import seaborn as sns
import os
import akerbp.models.rule_based_models.crossplot_helpers as crossplot_helpers


def _make_col_dtype_lists(df: DataFrame) -> Tuple[List, List]:
    """
    Returns lists of numerical and categorical columns

    Args:
        df (DataFrame): dataframe with columns to classify

    Returns:
        tuple: lists of numerical and categorical columns
    """
    num_cols = set(df._get_numeric_data().columns)
    cat_cols = list(set(df.columns) - set(num_cols))
    return list(num_cols), cat_cols


def _apply_metadata(df: DataFrame, **kwargs) -> Tuple[DataFrame, List, List]:
    """
    Applies specified metadata to data

    Args:
        df (DataFrame): dataframe to apply metadata to

    Returns:
        tuple: dataframe after applying metadata, list of numerical columns
        and list of categorical columns
    """
    num_cols, cat_cols = _make_col_dtype_lists(df)
    num_filler = kwargs.get("num_filler", None)
    cat_filler = kwargs.get("cat_filler", None)
    if num_filler is not None:
        df.loc[:, num_cols] = df[num_cols].replace(to_replace=num_filler, value=np.nan)
    if cat_filler is not None:
        df.loc[:, cat_cols] = df[cat_cols].replace(to_replace=cat_filler, value=np.nan)
    return df, num_cols, cat_cols


def _validate_features(X_pred_features: List, expected_curves: List):
    """
    Checks that provided and expected features are the same

    Args:
        X_pred_features (List): provided features list
        expected_curves (List): expected feature list

    Raises:
        ValueError: raises an error if expected and provided are different
    """
    if not X_pred_features.issubset(expected_curves):
        missing_features = list(expected_curves - X_pred_features)
        if len(missing_features) > 0:
            raise ValueError(
                f"Following features are expected but missing: {missing_features}"
            )
    if not expected_curves.issubset(X_pred_features):
        missing_features = list(X_pred_features - expected_curves)
        if len(missing_features) > 0:
            raise ValueError(
                f"Following features are expected but missing: {missing_features}"
            )


def _create_features(X_pred: DataFrame) -> DataFrame:
    """
    Creates features necessary for algorithms

    Args:
        X_pred (DataFrame): dataframe with rovided data of one well

    Returns:
        DataFrame: df with added cols of required data
    """
    if "VP" not in X_pred.columns:
        X_pred.loc[:, "VP"] = 304.8 / X_pred["AC"]
    if "VS" not in X_pred.columns:
        X_pred.loc[:, "VS"] = 304.8 / X_pred["ACS"]
    if "AI" not in X_pred.columns:
        X_pred.loc[:, "AI"] = X_pred["DEN"] * ((304.8 / X_pred["AC"])**2)
    if "VPVS" not in X_pred.columns:
        X_pred.loc[:, "VPVS"] = X_pred["VP"] / X_pred["VS"]
    return X_pred


def fill_holes(df_: DataFrame, flag_col: str, limit: int = 3) -> List:
    """
    Fill holes/include adjacent points to anoamlies as anomalies

    Args:
        df_ (DataFrame): dataframe with the column to fill holes
        flag_col (str): column of df to fill holes
        limit (int, optional): how many samples to include as anomalye adjacent
        to anomalies. Defaults to 3.

    Returns:
        list: list of new values
    """
    tmp = df_.copy()
    tmp["filled"] = tmp[flag_col].replace(
        0, np.nan
    ).fillna(
        method="ffill",
        limit=limit
    ).fillna(
        method="bfill",
        limit=limit
    ).replace(np.nan, 0)
    return tmp.filled.values


# FIXME! move to crossplot_helpers
def get_crossplot_scores(
    df_well: DataFrame,
    logname: str,
    curves: List,
    y_pred: DataFrame,
    method: str,
    **algo_params
) -> DataFrame:
    """
    Returns scores for each sample

    Args:
        df_well (DataFrame): dataframe with data from one well
        logname (str): log to which all curves will the crossploted against
        curves (list): list of curves to analyse
        y_pred (DataFrame): dataframe with results per sample
        method (str): crossplot type string

    Returns:
        DataFrame: y_pred with extra columns of scores
    """
    for y in curves:
        _, scores, idx = crossplot_helpers.find_crossplot_scores(
            df_well, x=logname, y=y, **algo_params
        )
        for a_method in scores.keys():
            y_pred.loc[
                idx, f"{a_method}_{method}_{logname.lower()}"
            ] = scores[a_method]
            y_pred[f"{a_method}_{method}_{logname.lower()}"].fillna(0, inplace=True)
    return y_pred


# FIXME! move to crossplot_helpers
def get_crossplot_anomalies(
    df_well: DataFrame, logname: str, curves: List, **algo_params
) -> List:
    """
    Returns bool for each sample indicating anomaly or not

    Args:
        df_well (DataFrame): dataframe with data from one well
        logname (str): log to which all curves will the crossploted against
        curves (list): list of curves to analyse

    Returns:
        DataFrame: y_pred with extra columns of scores
    """
    anomalies = []
    for y in curves:
        tmp, _, _ = crossplot_helpers.find_crossplot_scores(
            df_well, x=logname, y=y, **algo_params
        )
        anomalies.extend(tmp)
    return anomalies


def expand_flags(
    df_well: DataFrame, flag_col: str, expansion_size: int = 3
) -> DataFrame:
    """
    Expands the flagged samples by expansion_size in each direction

    Args:
        df_well (DataFrame): The input dataframe
        flag_col (str): Name of the curve to expand
        expansion_size (int, optional): how much to expand. Defaults to 3.

    Returns:
        df_well (DataFrame): input dataframe with updated flag_col
    """
    # get all available indices in the well dataframe
    well_idx = df_well.index.tolist()
    # then the flagged rows
    flag_idx = df_well[df_well[flag_col] == True].index.tolist()
    # Then find the flagged indices in the well
    flag_idx_idx = [well_idx.index(i) for i in flag_idx]
    for i in range(1, expansion_size):
        # check that the before and after indices are available in the well
        before_flag_idx_i = [
            well_idx[x - i]
            for x in flag_idx_idx if ((x - i) >= 0) and (well_idx[x - i] in well_idx)
        ]
        after_flag_idx_i = [
            well_idx[x + i]
            for x in flag_idx_idx
            if ((x + i) < len(well_idx)) and (well_idx[x + i] in well_idx)
        ]
        # update the flag_col
        df_well.loc[before_flag_idx_i, flag_col] = 1
        df_well.loc[after_flag_idx_i, flag_col] = 1
    return df_well


def print_metrics(
    true: Series,
    pred: Series,
    ax: List = None,
    print_values: bool = False,
    title: str = None,
    fig_name: str = None,
    plot_dir: str = None
):
    """
    Plot confusion matrix with metrics of bad logs detection

    Args:
        true (Series): true values
        pred (Series): predicted values
        ax (list, optional): axes indices. Defaults to None.
        print_values (bool, optional): specify if printing is required with confusion matrix. Defaults to False.
        title (str, optional): specifies title for confusion matrix. Defaults to None.
    """
    recall = sklearn.metrics.recall_score(true, pred)
    prec = sklearn.metrics.precision_score(true, pred)
    f1sc = sklearn.metrics.f1_score(true, pred)

    conf_matrix = sklearn.metrics.confusion_matrix(true, pred)

    LABELS = [False, True]
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(5, 5))
    sns.heatmap(
        conf_matrix,
        xticklabels=LABELS,
        yticklabels=LABELS,
        annot=True,
        annot_kws={"fontsize" : 14},
        fmt="d",
        ax=ax
    )
    if title is None:
        title = "Confusion matrix"
    if print_values:
        title = f"{title} \n Recall = {recall:.3f} \n Precision = {prec:.3f} "\
            f"\n F1-score = {f1sc:.2f}"
    ax.set_title(title, {"size" : "15"})
    ax.set_ylabel("Label", {"size" : "18"})
    ax.set_xlabel("Predicted class", {"size" : "18"})
    plt.tight_layout()
    if plot_dir is not None:
        plt.savefig(os.path.join(plot_dir, f"{fig_name}.jpg"), dpi=150)



from typing import Dict
import os
import numpy as np
import pandas as pd

from akerbp.models.classification_models import ClassificationModel
from akerbp.models.classification_models import XGBoostClassificationModel


class NestedShaleModel(ClassificationModel):
    """
    A wrapper object for two models
    1. A model that predicts shale
    2. A model that predicts all classes except shale
    """

    def __init__(self, shale_settings: Dict, other_settings: Dict, model_path: str):
        self.model_path = model_path
        self.shale_settings = shale_settings
        self.other_settings = other_settings
        self.shale_model = XGBoostClassificationModel(
            shale_settings, os.path.join(self.model_path, "shale_model")
        )
        self.other_model = XGBoostClassificationModel(
            other_settings, os.path.join(self.model_path, "other_model")
        )
        self.label_shale = 65000

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Returns predictions of shale

        Args:
            X (pd.DataFrame): dataframe with data to predict

        Raises:
            ValueError: raises an error if no model exists

        Returns:
            pd.DataFrame: dataframe with shale predictions
        """
        if self.shale_model is None or self.other_model is None:
            raise ValueError(
                "Model must be trained or loaded from file before predict can be run"
            )
        shale_prediction = self.shale_model.predict(X)
        other_prediction = self.other_model.predict(X[shale_prediction == -1])
        full_prediction = np.zeros(X.shape[0])
        full_prediction[shale_prediction == self.label_shale] = self.label_shale
        full_prediction[shale_prediction == -1] = other_prediction
        return full_prediction.astype(int)

    def train(self, df: pd.DataFrame, targets: pd.Series):
        """
        Train the binary shale classification model

        Args:
            df (pd.DataFrame): dataframe with input data to train models
            targets (pd.Series): series with label data
        """
        print("Training shale model..")
        targets_shale = np.copy(targets)
        targets_shale[targets_shale != self.label_shale] = -1
        self.shale_model.train(df, targets_shale)
        print("Training model for non-shale lithologies..")
        targets_nonshale = targets[targets != self.label_shale].copy()
        df_nonshale = df[targets != self.label_shale]
        self.other_model.train(df_nonshale, targets_nonshale)

    def save(self, model_path: str = None):
        """
        Saves models to given path

        Args:
            model_path (str, optional): location to save model to. Defaults to None.
        """
        if model_path is None:
            model_path = self.model_path
        self.shale_model.save(model_path=os.path.join(self.model_path, "shale_model"))
        self.other_model.save(model_path=os.path.join(self.model_path, "other_model"))



from typing import List
import pandas as pd
from pandas.core.frame import DataFrame, Series
import numpy as np


def keepNLargestGroups(df_: pd.DataFrame, n: int) -> List:
    """
    Returns flatlines indices for AC, ACS and DEN

    Args:
        df_ (Series): series with flags
        n (int): number of large groups of consecutive anomalies to keep

    Returns:
        List: processed flags
    """
    df = pd.DataFrame(df_.astype(int).values, columns=['flatline'])
    df['groups'] = df['flatline'].ne(df['flatline'].shift()).cumsum()
    count_groups = df[df.flatline == 1].groups.value_counts()
    count_groups = count_groups[:n].index.tolist()
    df['large_group'] = False
    df.loc[df.groups.isin(count_groups), 'large_group'] = True
    return df.large_group.values


def detect_flatlines(
    df: pd.DataFrame,
    col: str,
    lim: float,
    window: int = 5,
    n_largest_groups: int = 10
) -> List:
    """
    Returns flatlines indices given log

    Args:
        df (DataFrame): dataframe with log to be processed(col) and depth
        col (str): column to which find flags
        lim (float): threshold for consedeiring anomalous derivatives
        window (int): size of window to compared single point to
        n_largest_groups (int): number of large groups of consecutive anomalies to keep

    Returns:
        List: flags flor given column
    """

    # get sampling rate
    vc = df.DEPTH.diff().value_counts(normalize=True)
    sampling_rate = (vc * vc.index).sum()
    # get second derivative
    df['d2'] = df[col].diff().diff().abs() / sampling_rate
    # detect flatlines based on derivative
    df['d2'] = ((df['d2'] < lim).multiply(
        df['d2'].rolling(window, center=True).median().isnull())
    ) | (
        df['d2'].rolling(window, center=True).median() < lim
    )
    # get largest groups to avoid false positives
    df['d2'] = keepNLargestGroups(df['d2'], n_largest_groups)
    df['d2'] = np.where(df['d2'], 1, 0)
    return df['d2'].values  # which values are correct, d2 or v2? is v2 coming from d2?


def flag_flatline(
    df_well: DataFrame, y_pred: DataFrame = None, **kwargs
) -> DataFrame:
    """
    Returns flatlines indices for AC, ACS and DEN

    Args:
        df_well (DataFrame): data from one well
        y_pred (DataFrame): results. Defaults to None.

    Returns:
        DataFrame: df with added column of results
    """
    print("Method: flatline...")
    default_kwargs = {
        "n_largest_groups": 10,
        "den_lim": 0.01,
        "ac_lim": 0.005,
        "acs_lim": 0.01
    }
    user_kwargs = kwargs.get("flatline_params", {})
    kwargs = {}
    for k, v in default_kwargs.items():
        kwargs[k] = user_kwargs.get(k, v)

    if y_pred is None:
        y_pred = df_well.copy()

    y_pred.loc[:, [
        "flag_flatline_gen",
        "flag_flatline_ac",
        "flag_flatline_acs",
        "flag_flatline_den"
    ]] = 0, 0, 0, 0

    for col in ["AC", "ACS", "DEN"]:
        y_pred.loc[:, [f"flag_flatline_{col.lower()}"]] = detect_flatlines(
            df_well[["DEPTH", col]],
            col,
            kwargs[f"{col.lower()}_lim"],
            n_largest_groups=kwargs["n_largest_groups"]
        )
        y_pred.loc[:, "flag_flatline_gen"] =\
            y_pred["flag_flatline_gen"] | y_pred[f"flag_flatline_{col.lower()}"]
    return y_pred



from typing import Dict, Any
import os
import joblib
import xgboost
import pandas as pd
import numpy as np

from akerbp.models import Model


class RegressionModel(Model):
    pass


class XGBoostRegressionModel():
    """
    Subclass of akerbp.model

    Attributes
    ----------
        settings: dict
            the required keys for the dictionary:
                arbitrary model name: dict containing following keys (dict):
                    'model_parameters': dict of parameters required by xgboost.XGBRFClassifier (dict),
                    'use_class_weights': whether to use class weights in the model (boolean)
        model_path: str
            - If model_path is a path/to/dir/ then the serialized model is assumed to be (or will be saved to)
            path/to/dir/model.joblib, and the labedl encoder at /path/to/dir/le.joblib
            - If model_path is a full/path/to/my_model.joblib,
            then the label encoder is saved (or assumed to be) at /full/path/to/my_model_le.joblib

    Methods
    -------
        predict(X:dict/pandas.DataFrame)

        predict_proba(X:dict/pandas/DataFrame)

        train(df:pandas.DataFrame, targets:np.array)

        evaluate(X_test:pandas.DataFrame, y_test:pandas.DataFrame, metrics:list, label_map:dict, **kwargs)

        save(model_path:str)
    """
    def __init__(self, settings: Dict, model_path: str, load: bool = True):
        self.settings = settings
        for key, val in settings.items():
            setattr(self, key, val)
        self.model_path = model_path
        self._handle_model_path()
        if load:
            try:
                self.model = joblib.load(self.model_file_path)
                print("Model successfully loaded from ", self.model_file_path)
            except Exception as e:
                print(e)

    def _handle_model_path(self):
        """
        Generates model path if it does not yet exist
        """
        if os.path.isfile(self.model_path):
            self.model_file_path = self.model_path
        else:
            if not os.path.isdir(self.model_path):
                os.makedirs(self.model_path)
            self.model_file_path = os.path.join(self.model_path, "model.joblib")

    def _validate_features(self):
        """
        This function should be able to take in features in their
        raw, unprocessed form as read from the file test.csv and
        return predictions as an array integers of the same length

        Raises:
            ValueError: raises an error if provided features are not the expected ones

        Returns:
            [type]: [description]
        """
        if self.X_pred is None:
            return []
        # Check that the features in the model match the features provided in the data
        # (including their order)
        self.model_features = self.model.get_booster().feature_names
        if sorted(self.model_features) != sorted(self.X_pred.columns.values):
            raise ValueError(
                "Features in the provided dataset does not match the model expectation"
                f" - Expected: {list(self.model_features)}, "
                f"Provided: {list(self.X_pred.columns)}"
            )
        self.X_pred = self.X_pred[self.model_features]

    def predict(self, X: Any) -> np.ndarray:
        """
        Predicts based on given model

        Args:
            X (Any): datafrae or dictionary to predict with given model

        Raises:
            ValueError: raises an error if X not a datfarame or dict

        Returns:
            np.array: predictions array
        """
        if isinstance(X, dict):
            self.X_pred = pd.DataFrame.from_dict(X)
        elif isinstance(X, pd.core.frame.DataFrame):
            self.X_pred = X
        else:
            raise ValueError("Please pass the data as a dict or a pandas DataFrame")
        self._validate_features()
        return self.model.predict(self.X_pred, validate_features=True)

    def train(self, df: pd.DataFrame, targets: pd.Series):
        """
        Trains on the data given in the provided dataframe and targets.
        The curves of the model must be present in the dataframe.

        Args:
            df (pd.DataFrame): Pandas dataframe of features
            targets (pd.Series): List of integers of the same length as the number of
            rows in df
        """
        self.model = xgboost.XGBRegressor(**self.model_parameters)
        self.model.fit(df, targets)

    def explain(self):
        """
        Placeholder
        """
        raise ValueError("Not implemented for this model")

    def evaluate(self, X_test, y_test, metrics=None, label_map=None, **kwargs):
        """
        Placeholder
        """
        raise ValueError("Not implemented for this method")

    def save(self, model_path: str = None) -> str:
        """
        Saves model to given path

        Args:
            model_path (str, optional): path to where store model. Defaults to None.

        Returns:
            str: model path
        """
        if model_path is not None:
            self.model_path = model_path
        self._handle_model_path()
        joblib.dump(self.model, self.model_file_path)
        print("Model successfully saved to {}".format(self.model_file_path))
        return self.model_file_path
